# flashtool.py

import click
from flashtool.products.kmb.kmbpcie import KMBPCIeManager
from flashtool.products.kmb.srbpcie import SRBPCIeManager
from flashtool.products.kmb.xlinkwrapper import XlinkWrapperWinManager
from flashtool.thirdparty.fastboot import Fastboot
from flashtool.thirdparty.xlink import XlinkDependencyError
from flashtool.products.kmb.kmbusb import KMBUSBManager
from flashtool.products.kmb.kmberrors import NoDevicesFoundError
from flashtool.products.kmb.kmbimages import image_check, fip_check
import flashtool.helpers.logging as logging
import flashtool.helpers.configparser as configparser
import os
import platform


def cli_setup():
    app_default_cfg = {"app": {'force': False}}
    cp = configparser.ConfigParser(config_type="json")
    cp.read_dict(app_default_cfg)
    if platform.system() == "Linux":
        cp.read(os.path.expanduser('~/flashtool/config.json'))
    cp.read('config.json')
    return cp, logging.get_logger()


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version="2.5.0")
def cli():
    pass


@cli.command("flashall", short_help='flash all images into device. Type "flashtool flashall -h" for more.')
@click.option("-v", "--verbose", "verbose", count=True)
@click.option("-f", "--force", is_flag=True, help="Skip Image checking.")
@click.option("-d", '--device-type', type=click.Choice(('usb', 'pcie'), case_sensitive=False), multiple=True, help="Filter devices based on type: USB or PCIe")
@click.option("-s", "--slot", default=None, help="Select device in this slot ID. Use devices command to get slot ID.")
@click.argument("images", type=click.Path(exists=True, resolve_path=True), nargs=-1, required=True)
@click.option("--skip-format", is_flag=True, default=False, help="Skip GUID Partition Table creation before flashing the OS images.")
@click.option("--skip-secondary", is_flag=True, default=False, help="Do not flash secondary partition.")
@click.option("--skip-boot", is_flag=True, default=False, help="Do not continue to boot once flashed the images.")
@click.option("--wipe", is_flag=True, default=False, help="Wipe environment data.")
@click.option("--max-threads", is_flag=True, default=False, help="Create maximum threads to flash multiple devices concurrently")
def flashall(verbose, force, device_type, slot, images, skip_format, skip_secondary, skip_boot, wipe, max_threads):
    """Flash FIP and OS images. The flow differs according to the FIP file name. The FIP filenames supported are 'fip-evm.bin', 'fip-hddl2.bin', 'fip-m2.bin'\n
    The OS image files must be named according to their content like this: 'boot.img', 'system.img', 'syshash.img' and 'data.img'.\n
    If FIP and OS images are provided, the flow will be: Flash FIP-->Format device-->Flash primary and secondary partitions-->Continue to boot.
    The device needs to be in "FIP Recovery Mode"\n
    if only OS images provided, the flow will be: Format device-->Flash primary and secondary partitions-->Continue to boot.
    The device needs to be in "OS Recovery Mode"\n
    IMPORTANT: For EVM-PCIe board, only OS images can be flashed using this command.\n
    Flags:\n
    The flow can be selectively changed with --skip-format, --skip-secondary and --skip-boot flags.\n
    User can use --wipe flag to wipe uboot environment primary and secondary partitions. However, this option is not supported in Windows yet.\n
    For multiple devices, they will be flashed 3 devices concurrently at a time. --max-threads flag will enable all devices to be flashed concurrently.\n
    \n
    For other use cases, please refer to README.md in the flashtool package.
    """
    cp, logger = cli_setup()
    try:
        if verbose > 0:
            for handler in logger.root.handlers:
                handler.setLevel("DEBUG")
            logger.debug("VERBOSE On")

        kwargs = {}
        for path in images:
            filename = os.path.basename(path)
            if filename not in ["fip-evm.bin", "fip-m2.bin", "fip-hddl2.bin", "boot.img", "system.img", "syshash.img", "data.img", "factory.img"] and 'factory_' not in filename and 'fip-' not in filename:
                raise click.BadParameter("File name does not follow convention. The filenames supported are 'fip-*.bin', 'boot.img', 'system.img', 'syshash.img', 'data.img' and 'factory.img' or 'factory_[deviceid].img'.")
            else:
                if filename == "fip-hddl2.bin" or filename == "fip-m2.bin" or filename == "fip-evm.bin" or "fip-" in filename:
                    kwargs['fip'] = path
                elif 'factory_' in filename:
                    if 'factory' not in kwargs:
                        kwargs['factory'] = []
                    kwargs['factory'].append(path)
                else:
                    kwargs[filename[:-4]] = path

        if force:
            cp['app']['force'] = force
        if cp['app']['force']:
            logger.warning("Force mode on. Image checking skipped")
        else:
            for imgtype, img in kwargs.items():
                if imgtype == "fip":
                    fip_check(img)
                    logger.info("{} is a valid FIP".format(img))
                else:
                    if type(img) is list:
                        for im in img:
                            magic = image_check(im)
                            logger.info("{} is a valid {} image".format(im, magic))
                    else:
                        magic = image_check(img)
                        logger.info("{} is a valid {} image".format(img, magic))

        kwargs['skip_format'] = skip_format
        kwargs['skip_secondary'] = skip_secondary
        kwargs['skip_boot'] = skip_boot
        kwargs['wipe'] = wipe
        kwargs['max_threads'] = max_threads

        if len(device_type) > 1:
            raise click.BadOptionUsage(device_type, "--device-type can be provided only once")
        if len(device_type) == 0:
            device_type = ("usb", "pcie")
        if "usb" in device_type:
            if platform.system() == "Windows":
                logger.warning("USB flashing is not supported for Windows yet")
            else:
                logger.debug("USB devices selected")
                dm = KMBUSBManager()
                if slot:
                    dm.slot = slot
                try:
                    dm.flash_all(**kwargs)
                except NoDevicesFoundError as e:
                    logger.warning(str(e))
        if "pcie" in device_type:
            logger.debug("PCIE devices selected")
            fip = kwargs.get('fip', None)
            try:
                if fip:
                    if platform.system() == "Windows" and  os.path.basename(fip) == "fip-m2.bin" :
                        dm = XlinkWrapperWinManager()
                    else:
                        dm = SRBPCIeManager()
                else:
                    dm = KMBPCIeManager()
                if slot:
                    dm.slot = slot
                dm.flash_all(**kwargs)
            except XlinkDependencyError:
                logger.warning("Unable to find or flash PCIe devices due to missing drivers. If you intend to flash a PCIe device, please install the drivers and rerun the tool")
            except click.BadParameter as e:
                logger.warning(str(e))
            except NoDevicesFoundError as e:
                logger.warning(str(e))

    except Exception as ex:
        logger.exception("{}: {}".format(ex.__class__.__name__, str(ex)))


@cli.command("recover", short_help='flash FIP on EVM PCIe board only. Type "flashtool recover -h" for more.')
@click.option("-v", "--verbose", "verbose", count=True)
@click.option("-f", "--force", is_flag=True, help="Skip FIP checking.")
@click.option("-d", '--device-type', type=click.Choice(('usb', 'pcie'), case_sensitive=False), multiple=True, help="Filter devices based on type: USB or PCIe")
@click.option("-s", "--slot", default=None, help="Select device in this slot ID. Use devices command to get slot ID.")
@click.option("--max-threads", is_flag=True, default=False, help="Create maximum threads to flash multiple devices concurrently")
@click.argument("fip", type=click.Path(exists=True, resolve_path=True), nargs=1)
def recover(verbose, force, device_type, slot, max_threads, fip):
    """Flash FIP image. Can be used only on EVM PCIe board. \n
    The filename supported is 'fip-evm.bin'.\n
    For multiple devices, they will be flashed 3 devices concurrently at a time. --max-threads flag will enable all devices to be flashed concurrently.\n
    In order to flash OS images on the EVM PCIe board, please use 'flashall' command.\n
    \n
    For example, please refer to README.md in the flashtool package.
    """
    cp, logger = cli_setup()
    try:
        if verbose > 0:
            for handler in logger.root.handlers:
                handler.setLevel("DEBUG")
            logger.debug("VERBOSE On")

        filename = os.path.basename(fip)

        if len(device_type) > 1:
            raise click.BadOptionUsage(device_type, "--device-type can be provided only once")
        if len(device_type) == 0:
            device_type = ("usb", "pcie")

        if force:
            cp['app']['force'] = force
        if cp['app']['force']:
            logger.warning("Force mode on. FIP checking skipped")
        else:
            fip_check(fip)
            logger.info("{} is a valid FIP".format(fip))

        if "usb" in device_type:
            logger.warning("USB flashing is not supported. Please use flashall command")
        if "pcie" in device_type:
            logger.debug("PCIE devices selected")
            try:
                if platform.system() == "Windows" and os.path.basename(fip) == "fip-evm.bin":
                    dm = XlinkWrapperWinManager()
                else:
                    dm = KMBPCIeManager()
                if slot:
                    dm.slot = slot
                dm.recover(fip, max_threads=max_threads)
            except XlinkDependencyError:
                logger.warning("Unable to find or flash PCIe devices due to missing drivers. If you intend to flash a PCIe device, please install the drivers and rerun the tool")
            except NoDevicesFoundError as e:
                logger.warning(str(e))

    except Exception as ex:
        logger.exception("{}: {}".format(ex.__class__.__name__, str(ex)))


@cli.command("devices", short_help='List available devices. Type "flashtool devices -h" for more.')
@click.option("-v", "--verbose", "verbose", count=True)
@click.option("-d", '--device-type', type=click.Choice(('usb', 'pcie'), case_sensitive=False), multiple=True, help="Filter devices based on type: USB or PCIe")
def devices(verbose, device_type):
    """List Keembay devices. Optionally user can specify --device-type  to only view USB or PCIe devices"""
    cp, logger = cli_setup()
    try:
        if verbose > 0:
            for handler in logger.root.handlers:
                handler.setLevel("DEBUG")
            logger.debug("VERBOSE On")
        if not device_type:
            device_type = ('usb') # do only USB by default
        if 'usb' in device_type:
            if platform.system() == "Windows":
                logger.warning("USB flashing is not supported for Windows yet")
            else:
                logger.debug("USB devices selected")
                try:
                    for device in KMBUSBManager().devices:
                        logger.info(str(device))
                except NoDevicesFoundError as e:
                    logger.warning(str(e))
        if 'pcie' in device_type:
            logger.debug("PCIE devices selected")
            try:
                for device in KMBPCIeManager().devices:
                    logger.info(str(device))
            except XlinkDependencyError:
                logger.warning("Unable to find or flash PCIe devices due to missing drivers. If you intend to find or flash a PCIe device, please install the drivers and rerun the tool")
            except NoDevicesFoundError as e:
                logger.warning(str(e))
    except Exception as ex:
        logger.exception("{}: {}".format(ex.__class__.__name__, str(ex)))

@cli.command("fastboot", short_help='Call included fastboot binary directly', context_settings={"ignore_unknown_options": True})
@click.argument('args', nargs=-1)
def fastboot(args):
    _, ret = Fastboot().args(*args)
    print(ret.stdout.decode('utf-8', "ignore"), end='')

if __name__ == "__main__":
    cli()
