#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, Intel Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#

import threading
import itertools
import time
import sys
import concurrent.futures
import os
import fnmatch
import flashtool.helpers.logging as logging

class AsyncError(Exception):
    def __init__(self, error_list, completed_list=None):
        self.error_list = error_list
        self.completed_list = completed_list

    def __str__(self):
        error_list = '\n'.join(self.error_list)
        if self.completed_list:
            completed_list = '\n'.join(['{}:{}'.format(x, y) for x, y in self.completed_list])
            return "Threads reported following errors:\n{}\nFollowing threads completed:\n{}".format(error_list, completed_list)
        return "Threads reported following errors:\n{}".format(error_list)


class Signal:
    go = True
    playing = True
    pause = False


def spin(msg, signal):
    write, flush = sys.stdout.write, sys.stdout.flush
    for char in itertools.cycle("|/-\\"):
        if signal.playing:
            status = char + " " + msg
            write(status)
            flush()
            write("\x08" * len(status))
            time.sleep(0.1)
            if signal.pause:
                write(" " * len(status) + "\x08" * len(status))
                signal.playing = False
        if not signal.go:
            break
    write(" " * len(status) + "\x08" * len(status))


def async_dev_ops(devices, func_name, *args, spinner_msg=" Downloading..", max_threads=False):
    error_list = []
    results = {}

    # TODO(themarpe) - ignore max threads flag, and go all the way up to 32+5
    MAX_THREADS=32
    n = MAX_THREADS + 5 # (Some are reserved for IO bound ops)
    # Raise an exception if there are more devices than threads as it causes issues
    if len(devices) > MAX_THREADS:
        raise AsyncError('Too many devices, reduce number of devices to continue.', list(devices))

    iters = [iter(devices)] * n
    device_groups = itertools.zip_longest(*iters, fillvalue=None)
    for devices in device_groups:
        with concurrent.futures.ThreadPoolExecutor(max_workers=n) as executor:
            # Start the load operations and mark each future with its device
            if len(args) > 0:
                future_device_dict = {executor.submit(getattr(device, func_name), *args): device for device in devices if device is not None}
            else:
                future_device_dict = {executor.submit(getattr(device, func_name)): device for device in devices if device is not None}
            signal = Signal()
            spinner = threading.Thread(target=spin, args=(spinner_msg, signal))
            spinner.start()
            for future in concurrent.futures.as_completed(future_device_dict):
                device = future_device_dict[future]
                try:
                    data = future.result()
                except Exception as exc:
                    error = '{} generated an exception: {}'.format(device, exc)
                    error_list.append(error)
                else:
                    results[device] = data
            signal.go = False
            spinner.join()
    if len(error_list) > 0:
        # Only warn if this fip flashing fails, the check happens afterwards that all same devices are in OS mode afterwards
        if func_name == 'flash_fip':
            logging.get_logger().warning(f'Issue flashing FIP: {error_list}')
        else:
            raise AsyncError(error_list, list(results.items()))
    return results


def find_dir(pattern, path):
    for root, dirs, files in os.walk(path):
        for name in dirs:
            if fnmatch.fnmatch(name, pattern):
                return os.path.join(root, name)
    return None

def find_any(pattern, path):
    for root, dirs, files in os.walk(path):
        for name in dirs:
            if fnmatch.fnmatch(name, pattern):
                return os.path.join(root, name)
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                return os.path.join(root, name)
    return None
