#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, Intel Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#

import logging
import click


class ClickHandler(logging.StreamHandler):
    '''This class colors the console output'''
    # levelcolor = {
    #     "CRITICAL": "red",
    #     "ERROR": "red",
    #     "WARNING": "yellow",
    #     "INFO": None,
    #     "DEBUG": None,
    #     "NOTSET": None,
    #     "SUCCESS": "green"
    #     }

    levelcolor = {
        50: "red",
        40: "red",
        30: "yellow",
        20: "green",
        10: None,
        0: None,
        }

    def __init__(self, stream=None):
        super().__init__(stream)

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to console using click.echo(). Foregorund
        color is chosen according to level and set using click.style(msg, fg).
        """
        try:
            msg = self.format(record)
            fg = self.levelcolor.get(record.levelno, None)
            click.echo(click.style(msg, fg=fg))
        except Exception:
            self.handleError(record)


class SuppressExceptionFormatter(logging.Formatter):
    '''Supresses the Traceback of error'''
    def format(self, record):
        """
        Format the specified record as text.

        The record's attribute dictionary is used as the operand to a
        string formatting operation which yields the returned string.
        Before formatting the dictionary, a couple of preparatory steps
        are carried out. The message attribute of the record is computed
        using LogRecord.getMessage(). If the formatting string uses the
        time (as determined by a call to usesTime(), formatTime() is
        called to format the event time. If there is exception information,
        it is discarded.
        """
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        s = self.formatMessage(record)
        return s


def get_logger():
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler('flashtool.log')
    fh.setLevel(logging.DEBUG)
    # create console handler with a higher log level
    ch = ClickHandler()
    ch.setLevel(logging.INFO)
    # create formatter and add it to the handlers
    formatter_log = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    formatter_console = SuppressExceptionFormatter('%(message)s')
    ch.setFormatter(formatter_console)
    fh.setFormatter(formatter_log)
    # add the handlers to logger
    logger.addHandler(ch)
    logger.addHandler(fh)
    return logger
