#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, Intel Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#

import os


class FIPCheckError(Exception):
    def __init__(self, fip):
        self.fip = fip

    def __str__(self):
        return "FIP check failed for FIP:{}".format(self.fip)


class ImageCheckError(Exception):
    def __init__(self, image):
        self.image = image

    def __str__(self):
        return "Image check failed for image:{}".format(self.image)


class UnknownImageError(Exception):
    def __str__(self):
        return "Filename not supported."


class EmptyFileError(Exception):
    def __init__(self, image):
        self.image = image

    def __str__(self):
        return "Image:{} is an empty file".format(self.image)


def empty_check(image, **kwds):
    if os.stat(image).st_size == 0:
        raise EmptyFileError(image)


def fip_check(fip, **kwds):
    empty_check(fip)
    with open(fip, 'rb') as image:
        magic = image.read(4)
        if magic != b'\x01\x00\x64\xaa':
            raise FIPCheckError(fip)


def image_check(image, **kwds):
    magic_mapping = {
        'boot.img': 'boot',
        'system.img': 'system',
        'syshash.img': 'verity',
        'data.img': 'data',
        'factory.img': 'factory'
    }
    try:
        if 'factory_' in image:
            magic = 'factory'
        else:
            magic = magic_mapping[os.path.basename(image)]
        with open(image, 'rb') as f:
            x = 6
            if not magic == 'verity':
                f.seek(1144, 1)
                if magic == 'data' or magic == 'boot':
                    x = 4
                elif magic == 'factory':
                    x = 7
            if f.read(x).decode("ascii") == magic:
                return magic
            else:
                raise ImageCheckError(image)
    except KeyError:
        raise UnknownImageError from None
