#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, Intel Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#

import logging
import os
import shutil
import time
import platform
from subprocess import PIPE, run
from flashtool.common.utils import async_dev_ops, find_dir
from flashtool.thirdparty.xlink import Xlink
from flashtool.products.kmb.kmberrors import NoDevicesFoundError, DeviceModeError
from ctypes import c_uint32

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class _KMBPCIeDevice:
    status_dict = {
        0: "OFF",
        1: "Error",
        2: "Fip recovery",
        3: "OS recovery",
        4: "Ready",
    }

    def __init__(self, sw_id, name):
        self.logger = logger
        self.sw_id = sw_id
        self.name = name
        self._sysfs_path = None
        self._status = None
        self.xlink = Xlink()

    @property
    def sysfs_path(self):
        if self._sysfs_path is None:
            self._sysfs_path = find_dir('*{}*'.format(self.name), '/sys/devices')
        return self._sysfs_path

    @property
    def status(self):
        self._status = self.xlink.get_device_status(self.sw_id)
        return self.status_dict[self._status]

    def remove(self):
        self.xlink.remove(self)

    def boot_device(self, fip=None):
        self.xlink.boot_device(self.sw_id, fip)

    def flash_fip(self, fip=None):
        start = time.time()
        if self.status != "Fip recovery":
            raise DeviceModeError("Fip recovery", self)
        self.boot_device(fip)
        end = time.time()
        return end - start

    def flash_partition(self, part_name, img):
        start = time.time()
        self.xlink.flash_partition(self, part_name, img)
        end = time.time()
        return end - start

    def erase(self, part_name):
        start = time.time()
        self.xlink.erase(self, part_name)
        end = time.time()
        return end - start

    def create_gpt(self):
        start = time.time()
        self.xlink.create_gpt(self)
        end = time.time()
        return end - start

    def recovery_done(self):
        start = time.time()
        self.xlink.recovery_done(self)
        end = time.time()
        return end - start

    def __repr__(self):
        return "{} Intel Keembay PCIe(in mode:{})".format(self.name, self.status)


class KMBPCIeManager:
    def __init__(self, slot=None):
        self.logger = logger
        self.xlink = Xlink()
        self._devices = None
        self.slot = slot

    def _get_device_list(self):
        sw_id_list, num_dev = self.xlink.get_device_list()
        tuple_set = set()
        for i in range(num_dev):
            sw_device_id = sw_id_list[i]
            name = self.xlink.get_device_name(sw_device_id)
            tuple_set.add((sw_device_id, name))
        devices = [
            _KMBPCIeDevice(sw_device_id, name) for sw_device_id, name in tuple_set
        ]
        self.logger.debug("PCIe devices list:{}".format(devices))
        return devices

    @property
    def devices(self):
        if self._devices is None:
            self._devices = self._get_device_list()
        devices = [device for device in list(self._devices) if self.slot is None or device.name == self.slot]
        if devices:
            return devices
        else:
            raise NoDevicesFoundError(device_type='PCIe')

    def recover(self, fip, **kwargs):
        if platform.system() == "Windows":
            if not os.path.exists("C:\\intel-kmb\\"):
                os.mkdir("C:\\intel-kmb\\")
            shutil.copy(fip, "C:\\intel-kmb\\fip.bin")
        else:
            if not os.path.exists("/lib/firmware/intel-kmb"):
                os.mkdir("/lib/firmware/intel-kmb")
            shutil.copy(fip, "/lib/firmware/intel-kmb/fip.bin")
        self.logger.debug("Starting FIP flashing")
        for device in self.devices:
            elapsed = device.flash_fip()
            self.logger.info('{} completed FIP download in {:.4f}s'.format(device, elapsed))
        if "fip-m2.bin" in fip:
            self.logger.info(f'FIP flashed for {device.name}. Waiting for 30s to check status...')
            time.sleep(30)
            if platform.system() == "Windows":
                self.logger.warning("Please disable then enable \"Intel(R) KeemBay XLink PCIe Device\" in device manager and run './flashtool flashall IMAGES...' to continue OS images flashing")
            else:
                self.logger.warning("Please switch to Normal Mode then run './flashtool flashall IMAGES...' to continue OS images flashing")
        else:
            self.logger.warning("Please shutdown the system, set the boot switch to normal and reboot the system and run './flashtool flashall IMAGES...' to continue OS images flashing")

    def flash(self, part_name, img, **kwargs):
        max_threads = kwargs.get("max_threads", False)
        self.logger.debug("Starting flashing")
        for device, data in async_dev_ops(self.devices, 'flash_partition', part_name, img, spinner_msg="Flashing", max_threads=max_threads).items():
            self.logger.info('{} flashed in {:.4f}s'.format(device, data))

    def erase(self, part_name, **kwargs):
        self.logger.debug("Erasing {}".format(part_name))
        max_threads = kwargs.get("max_threads", False)
        for device, data in async_dev_ops(self.devices, 'erase', part_name, spinner_msg="Erasing {}".format(part_name), max_threads=max_threads).items():
            self.logger.info('{} erased {} in {:.4f}s'.format(device, part_name, data))

    def format(self, **kwargs):
        self.logger.debug("Formatting")
        max_threads = kwargs.get("max_threads", False)
        for device, data in async_dev_ops(self.devices, 'create_gpt', max_threads=max_threads).items():
            self.logger.info('{} formatted in {:.4f}s'.format(device, data))

    def boot(self, **kwargs):
        self.logger.debug("Signalling recovery done")
        max_threads = kwargs.get("max_threads", False)
        for device, data in async_dev_ops(self.devices, 'recovery_done', max_threads=max_threads).items():
            self.logger.debug('{} signalled recovery done in {:.4f}s'.format(device, data))
            self.logger.info("Flashing complete for device {}. Please shutdown the system, set the boot switch to normal and reboot the system".format(device))

    def flash_all(self, **kwargs):
        max_threads = kwargs.get("max_threads", False)

        for device in self.devices:
            device.boot_device()

        time.sleep(1)

        if not kwargs.get("skip_format", False):
            self.format(max_threads=max_threads)

        if kwargs.get("wipe", False):
            if platform.system() == "Windows":
                self.logger.warning("Erasing is not supported in Windows yet. Ignoring --wipe option")
            else:
                self.erase("env-main", max_threads=max_threads)
                self.erase("env-redund", max_threads=max_threads)

        partition_img_mapping = {
            "boot": "boot_a",
            "system": "system_a",
            "syshash": "syshash_a",
            "data": "data",
        }

        win_part_img_mapping = {
            "boot": c_uint32(0),
            "system": c_uint32(1),
            "syshash": c_uint32(2),
            "data": c_uint32(6),
        }
        win_file_name = {
            "boot": b"C:\\intel-kmb\\boot.img",
            "system": b"C:\\intel-kmb\\system.img",
            "syshash": b"C:\\intel-kmb\\syshash.img",
            "data": b"C:\\intel-kmb\\data.img",
        }

        win_image_name = {
            "boot": b"intel-kmb\\boot.img",
            "system": b"intel-kmb\\system.img",
            "syshash": b"intel-kmb\\syshash.img",
            "data": b"intel-kmb\\data.img",
        }
        secondary_mapping = {
            "boot": "boot_b",
            "system": "system_b",
            "syshash": "syshash_b",
        }

        win_secondary_mapping = {
            "boot": c_uint32(0),
            "system": c_uint32(1),
            "syshash": c_uint32(2),
        }

        for key, val in kwargs.items():
            try:
                if platform.system() == "Windows":
                    part_name = win_part_img_mapping[key]
                    shutil.copy(val, win_file_name[key])
                    image = win_image_name[key]
                else:
                    part_name = partition_img_mapping[key]
                    image = val
            except KeyError:
                continue
            self.flash(part_name, image, max_threads=max_threads)

            if not kwargs.get("skip_secondary", False):
                if platform.system() == "Windows":
                    try:
                        part_name = win_secondary_mapping[key]
                        self.flash(part_name, image, max_threads=max_threads)
                    except KeyError:
                        pass
                else:
                    try:
                        part_name = secondary_mapping[key]
                        self.flash(part_name, image, max_threads=max_threads)
                    except KeyError:
                        pass

        if not kwargs.get("skip_boot", False):
            self.boot(max_threads=max_threads)
