#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, Intel Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#

import os
import shutil
import time
import platform
import sys
from click import BadParameter
from ctypes import c_uint32
from flashtool.common.utils import async_dev_ops
from flashtool.products.kmb.kmbpcie import KMBPCIeManager, DeviceModeError
from flashtool.thirdparty.xlink import SysDeviceError


class SRBPCIeManager(KMBPCIeManager):

    def reenumerate(self):
        if platform.system() == "Windows":
            time.sleep(30)
            try:
                _ = input("Scan for hardware changes in device manager and press any key here to continue")
            except EOFError:
                sys.exit()
        else:
            devices = self._get_device_list()
            sysfs_parent_paths = {
                os.path.dirname(device.sysfs_path) for device in devices
            }
            for device in devices:
                device.remove()
                del(device)
            self._devices = None
            self.logger.debug("PCIe devices removed")
            time.sleep(30)
            for path in sysfs_parent_paths:
                fn = os.path.join(path, "rescan")
                try:
                    with open(fn, "w") as f:
                        f.write("1")
                except Exception as e:
                    raise SysDeviceError(fn, e) from None
            self.logger.debug("PCIe devices rescanned")
            return self.devices

    def recover(self, fip, **kwargs):
        max_threads = kwargs.get("max_threads", False)
        if platform.system() == "Windows":
            if not os.path.exists("C:\\intel-kmb\\"):
                os.mkdir("C:\\intel-kmb\\")
            shutil.copy(fip, "C:\\intel-kmb\\fip.bin")
            self.logger.debug("Starting FIP flashing")
            for device in self.devices:
                device.flash_fip()
                self.logger.info(f'FIP flashed for {device.name}. Waiting for 30s to check status...')
                time.sleep(30)
                _ = input("Scan for hardware changes in device manager and press any key here to continue")
                if device.status == "OS recovery":
                    self.logger.info("FIP flashed successfully for device {}".format(device))
        else:
            if not os.path.exists("/lib/firmware/intel-kmb"):
                os.mkdir("/lib/firmware/intel-kmb")
            shutil.copy(fip, "/lib/firmware/intel-kmb/fip.bin")
            self.logger.debug("Starting FIP flashing")
            for device, data in async_dev_ops(self.devices, 'flash_fip', max_threads=max_threads).items():
                self.logger.debug('{} completed FIP download in {:.4f}s'.format(device, data))
            self.logger.debug("Waiting 30s before reenumarating")
            self.reenumerate()
            for device in self.devices:
                for count in range(30):
                    if device.status == "OS recovery":
                        self.logger.info("FIP flashed successfully for device {}".format(device))
                        break
                    else:
                        if count == 29:
                            raise DeviceModeError("OS recovery", device)
                        else:
                            time.sleep(1)
                            self.logger.debug("Waited {}s for device {} to come up in OS recovery mode".format(count, device))

    def flash_all(self, **kwargs):
        max_threads = kwargs.get("max_threads", False)
        fip = kwargs.get("fip", None)
        if fip:
            if os.path.basename(fip) == "fip-evm.bin":
                raise BadParameter("For KMB EVM PCIe, please flash fip using recover function. Then remove recovery jumper and power cycle host. Once up, flash OS images using flashall function.")
            else:
                self.recover(fip)
        for device in self.devices:
            device.boot_device()

        time.sleep(1)

        if not kwargs.get("skip_format", False):
            self.format(max_threads=max_threads)

        if kwargs.get("wipe", False):
            if platform.system() == "Windows":
                self.logger.warning("Partition erasing is not enabled for Windows")
            else:
                self.erase("env-main", max_threads=max_threads)
                self.erase("env-redund", max_threads=max_threads)

        partition_img_mapping = {
            "boot": "boot_a",
            "system": "system_a",
            "syshash": "syshash_a",
            "data": "data",
        }

        win_part_img_mapping = {
            "boot": c_uint32(0),
            "system": c_uint32(1),
            "syshash": c_uint32(2),
            "data": c_uint32(6),
        }
        win_file_name = {
            "boot": b"C:\\intel-kmb\\boot.img",
            "system": b"C:\\intel-kmb\\system.img",
            "syshash": b"C:\\intel-kmb\\syshash.img",
            "data": b"C:\\intel-kmb\\data.img",
        }

        win_image_name = {
            "boot": b"intel-kmb\\boot.img",
            "system": b"intel-kmb\\system.img",
            "syshash": b"intel-kmb\\syshash.img",
            "data": b"intel-kmb\\data.img",
        }
        secondary_mapping = {
            "boot": "boot_b",
            "system": "system_b",
            "syshash": "syshash_b",
        }

        for key, val in kwargs.items():
            try:
                if platform.system() == "Windows":
                    part_name = win_part_img_mapping[key]
                    shutil.copy(val, win_file_name[key])
                    image = win_image_name[key]
                else:
                    part_name = partition_img_mapping[key]
                    image = val
            except KeyError:
                continue
            self.flash(part_name, image, max_threads=max_threads)

            if not kwargs.get("skip_secondary", False):
                if platform.system() == "Windows":
                    self.logger.warning("Secondary Partition flashing is not enabled for Windows")
                else:
                    try:
                        part_name = secondary_mapping[key]
                        self.flash(part_name, image, max_threads=max_threads)
                    except KeyError:
                        pass

        if not kwargs.get("skip_boot", False):
            self.boot(max_threads=max_threads)
