from subprocess import run, PIPE, STDOUT
import os
import logging
import platform
import re

cwd = os.path.dirname(__file__)
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
legacy_fastboot = False

def exe_cmd(*cmds, **kwds):
    """Executes commands in a new shell. Directing stderr to PIPE.
    This is fastboot's own exe_cmd because of its peculiar way of writing
    non-error info to stderr.
    Args:
        cmds: A sequence of commands and arguments.
    Returns:
        The output of the command run.
    Raises:
        Exception is raised if an error occurred during the command execution.
    """
    cmd = ' '.join(cmds)
    _ = kwds.setdefault("stdout", PIPE)
    _ = kwds.setdefault("stderr", STDOUT)
    _ = kwds.setdefault("cwd", cwd)
    _ = kwds.setdefault("shell", True)
    _ = kwds.setdefault("timeout", None)
    logger.debug("cmd:{}, kwds: {}".format(cmd, kwds))
    cp = run(cmd, **kwds)
    stdout = cp.stdout.decode("utf-8", "ignore")
    logger.debug("cmd:{}, stdout:{}".format(cmd, stdout))
    return stdout, cp


def output_parse(stdout):
    for line in stdout.split('\n'):
        if "FAILED" in line or "error" in line:
            raise FastbootError(stdout)
        else:
            if "Finished" in line:
                line = line[line.find(":") + 1:]
                time = float(line.strip(' ').rstrip('s'))
                return time


class FastbootError(Exception):
    """Raised when there is an error in fastboot operations."""
    def __init__(self, stdout):
        self.stdout = stdout

    def __str__(self):
        return self.stdout


class Fastboot():
    """Proxy class for fastboot.
    For syntactic reasons, the '-' in fastboot commands need to be replaced
    with '_'. Can directly execute fastboot commands on an object:
    >> fb = FastbootProxy(<serial>)
    >> fb.devices() # will return the console output of "fastboot devices".
    """
    def __init__(self, serial="", name=""):
        self.serial = serial
        self.name = name
        # TODO(thamapre) - 'usb:' is also a part of the name. 
        # 'devices()' could return with that prepended already
        # Just regex would have to be modified to match against eg usb/tcp/udp/...
        if legacy_fastboot:
            if name:
                self.fastboot_str = "./fastboot -s usb:{}".format(name)
            elif serial:
                self.fastboot_str = "./fastboot -s {}".format(serial)
            else:
                self.fastboot_str = "./fastboot"
        else:
            if name:
                self.fastboot_str = "./{}/fastboot -s usb:{}".format(platform.machine(), name)
            elif serial:
                self.fastboot_str = "./{}/fastboot -s {}".format(platform.machine(), serial)
            else:
                self.fastboot_str = "./{}/fastboot".format(platform.machine())

    def _exec_fastboot_cmd(self, *cmds, **kwds):
        fastboot_cmd = (self.fastboot_str, *cmds)
        return exe_cmd(*fastboot_cmd, **kwds)

    def args(self, *args, **kwds):
        return exe_cmd(' '.join((self.fastboot_str, ) + args), **kwds)

    def devices(self):
        cmds = ("devices", "-l")
        stdout, cp = self._exec_fastboot_cmd(*cmds, timeout=5)
        devices = []

        # Check if no permissions was printed out
        if 'no permissions;' in stdout:
            import textwrap
            permissions_commands = textwrap.dedent("""
            echo 'SUBSYSTEMS=="usb", ATTRS{idVendor}=="8087", ATTRS{idProduct}=="0b39", GROUP="plugdev"' | sudo tee /etc/udev/rules.d/80-keembay.rules
            echo 'SUBSYSTEMS=="usb", ATTRS{idVendor}=="8087", ATTRS{idProduct}=="da00", GROUP="plugdev"' | sudo tee -a /etc/udev/rules.d/80-keembay.rules
            sudo udevadm control --reload-rules && sudo udevadm trigger
            """)
            raise RuntimeError('No permissions while scanning for USB devices! Execute the following first:\n{}'.format(permissions_commands))

        try:
            # Use regex for more robust matching
            matches = re.findall(r'(\S*) [\S\s]+?(?=usb:)usb:(\S*)', stdout.strip())
            logger.debug(f'matches {matches}')
            for match in matches:
                try:
                    # More graceful parsing
                    serialno = match[0]
                    name = match[1]

                    # Try retrieving actual device serial
                    # Use device path instead of "partial serial" as those can clash
                    orig_serial = serialno
                    logger.debug(f'about to query device {serialno} {name}')
                    serialno_stdout, cp = self.args(f'getvar serialno -s usb:{name}', timeout=5)
                    serialno_var = serialno_stdout.split('serialno: ')[1].split('\n')[0].strip()
                    if serialno_var == serialno:
                        serialno = serialno_var
                    else:
                        # Transform device serial into consistent uint64_t -> %016X (16 hex upper case)
                        # USB ROM BL reports it as "little endian hex string", rotate the hex bytes around
                        rotated_serial = serialno_var
                        serialno = ''
                        while rotated_serial != "":
                            serialno = f'{serialno}{rotated_serial[-2:]}'
                            rotated_serial = rotated_serial[:-2]
                        serialno = serialno.upper()

                    logger.debug("orig_serial: {}, serialno_var: {}, serialno:{}".format(orig_serial, serialno_var, serialno))

                    devices.append((name, serialno))
                except ValueError as e:
                    logger.debug("Exception: {}".format(e))
                    pass
        except Exception as e:
            logger.debug("Exception: {}".format(e))
            pass
        return devices

    def getvar(self, var):
        cmds = ("getvar", var)
        stdout, cp = self._exec_fastboot_cmd(*cmds)
        for line in stdout.split('\n'):
            if 'FAILED' in line:
                if 'Variable not implemented' in line:
                    return 'Variable not implemented'
                else:
                    raise FastbootError(stdout)
            else:
                if var in line:
                    return line[line.find("{}: ".format(var)) + len("{}: ".format(var)):]

    def stage(self, img):
        cmds = ("stage", img)
        stdout, cp = self._exec_fastboot_cmd(*cmds)
        return output_parse(stdout)

    def format(self):
        cmds = ("oem", "format")
        stdout, cp = self._exec_fastboot_cmd(*cmds)
        return output_parse(stdout)

    def flash(self, part_name, img):
        cmds = ("flash", part_name, img)
        stdout, cp = self._exec_fastboot_cmd(*cmds)
        return output_parse(stdout)

    def erase(self, part_name):
        cmds = ("erase", part_name)
        stdout, cp = self._exec_fastboot_cmd(*cmds)
        return output_parse(stdout)

    def reboot(self):
        cmds = ("reboot",)
        stdout, cp = self._exec_fastboot_cmd(*cmds)
        return output_parse(stdout)
