from flashtool.common.bases import Singleton
from subprocess import run, PIPE, CalledProcessError
from ctypes import CDLL, Structure, c_uint32, c_int, byref, c_char, cdll
import os
import _ctypes
import platform
import shutil


class XlinkDependencyError(Exception):
    def __init__(self, component, error):
        self.component = component
        self.error = error

    def __str__(self):
        return """Failed to load {}. Have you installed the correct debian package?
            Error details:{} {}.""".format(self.component, self.error.output, self.error.stderr)


class XlinkAPIError(Exception):
    def __init__(self, api, rc, dev_id=None):
        self.api = api
        self.rc = rc
        self.dev_id = dev_id

    def __str__(self):
        error_str = "Xlink API {} error with return code {}".format(self.api, self.rc)
        if self.dev_id:
            error_str = "{} for device {}".format(error_str, self.dev_id)
        return error_str


class SysDeviceError(Exception):
    def __init__(self, fn, err, device=None):
        self.fn = fn
        self.err = err
        self.device = device

    def __str__(self):
        err_str = "sysdevice {} error due to {}".format(self.fn, self.err)
        if self.device:
            err_str = "{} in device {}".format(err_str, self.device)
        return err_str


class XlinkHandle(Structure):
    _fields_ = [
        ('sw_device_id', c_uint32),
        ('dev_type', c_int)
    ]


class Xlink(metaclass=Singleton):

    def __init__(self):
        self.setup()

    @property
    def binary_fip_path(self):
        if platform.system() == "Windows":
            return b"intel-kmb\\fip.bin"
        else:
            return b"intel-kmb/fip.bin"

    def setup(self):
        if platform.system() == "Windows":
            os.add_dll_directory(os.path.abspath(r'C:\Program Files\Intel\XLink-Keembay'))
            self._xlink = cdll.LoadLibrary("XLink_DLL.dll")
        else:
            try:
                run('lsmod | grep mxlk', shell=True, stdout=PIPE, stderr=PIPE, check=True)
            except CalledProcessError:
                try:
                    run('modprobe mxlk', shell=True, stdout=PIPE, stderr=PIPE, check=True)
                except CalledProcessError as cpe:
                    raise XlinkDependencyError('PCIe driver', cpe) from None
            try:
                run('lsmod | grep xlink', shell=True, stdout=PIPE, stderr=PIPE, check=True)
            except CalledProcessError:
                try:
                    run('modprobe xlink', shell=True, stdout=PIPE, stderr=PIPE, check=True)
                except CalledProcessError as cpe:
                    raise XlinkDependencyError('Xlink driver', cpe) from None
            try:
                run('ldconfig -p | grep libXLink.so', shell=True, stdout=PIPE, stderr=PIPE, check=True)
            except CalledProcessError as cpe:
                raise XlinkDependencyError('Xlink library', cpe) from None
            self._xlink = CDLL("libXLink.so")  # load libXLink.so
        rc = self._xlink.xlink_initialize()
        if rc != 0:
            raise XlinkAPIError('xlink_initialize', rc)

    def teardown(self):
        _ctypes.dlclose(self._xlink._handle)
        os.close(4)
        run('modprobe -r xlink', shell=True, stdout=PIPE, stderr=PIPE, check=True)
        run('modprobe -r mxlk', shell=True, stdout=PIPE, stderr=PIPE, check=True)
        self._xlink = None

    def _win_get_device_status(self, sw_id):
        status = c_uint32()
        handle = XlinkHandle()
        handle.dev_type = c_int(0)
        handle.sw_device_id = c_uint32(sw_id)
        rc = self._xlink.xlink_get_device_status(byref(handle), byref(status))
        if rc != 0:
            raise XlinkAPIError('xlink_get_device_status', rc, sw_id)
        return status.value

    def _linux_get_device_status(self, sw_id):
        status = c_uint32()
        handle = XlinkHandle()
        handle.sw_device_id = c_uint32(sw_id)
        handle.dev_type = c_int(1)
        rc = self._xlink.xlink_get_device_status(byref(handle), byref(status))
        if rc != 0:
            raise XlinkAPIError('xlink_get_device_status', rc, sw_id)
        return status.value

    def get_device_status(self, sw_id):
        if platform.system() == "Windows":
            return self._win_get_device_status(sw_id)
        else:
            return self._linux_get_device_status(sw_id)

    def _win_get_device_list(self):
        sw_id_list = (c_uint32 * 8)()
        num_dev = c_uint32(8)
        rc = self._xlink.xlink_get_device_list(byref(sw_id_list), byref(num_dev))
        if rc != 0:
            raise XlinkAPIError("xlink_get_device_list", rc)
        return list(sw_id_list), num_dev.value

    def _linux_get_device_list(self):
        sw_id_list = (c_uint32 * 32)()
        num_dev = c_uint32()
        pid = c_int(0x6240)
        rc = self._xlink.xlink_get_device_list(byref(sw_id_list), byref(num_dev), pid)
        if rc != 0:
            raise XlinkAPIError("xlink_get_device_list", rc)
        return list(sw_id_list), num_dev.value

    def get_device_list(self):
        if platform.system() == "Windows":
            return self._win_get_device_list()
        else:
            return self._linux_get_device_list()

    def _win_get_device_name(self, sw_device_id):
        handle = XlinkHandle()
        handle.dev_type = c_int(0)
        handle.sw_device_id = c_uint32(sw_device_id)
        name = (c_char * 128)()
        rc = self._xlink.xlink_get_device_name(byref(handle), byref(name), c_int(128))
        if rc != 0:
            raise XlinkAPIError("xlink_get_device_name", rc, sw_device_id)
        return name.value.decode('ASCII')

    def _linux_get_device_name(self, sw_device_id):
        handle = XlinkHandle()
        handle.dev_type = c_int(1)
        handle.sw_device_id = c_uint32(sw_device_id)
        name = (c_char * 32)()
        rc = self._xlink.xlink_get_device_name(byref(handle), byref(name), 32)
        if rc != 0:
            raise XlinkAPIError("xlink_get_device_name", rc, sw_device_id)
        return name.value.decode('ASCII')

    def get_device_name(self, sw_device_id):
        if platform.system() == "Windows":
            return self._win_get_device_name(sw_device_id)
        else:
            return self._linux_get_device_name(sw_device_id)

    def _win_boot_device(self, sw_id, fip=None):
        handle = XlinkHandle()
        handle.dev_type = c_int(0)
        handle.sw_device_id = c_uint32(sw_id)
        try:
            binary_fip_path = fip.encode('utf-8')
        except AttributeError:
            binary_fip_path = self.binary_fip_path
        rc = self._xlink.xlink_boot_device(byref(handle), binary_fip_path)
        return rc

    def _linux_boot_device(self, sw_id, fip=None):
        handle = XlinkHandle()
        handle.sw_device_id = c_uint32(sw_id)
        handle.dev_type = c_int(1)
        try:
            binary_fip_path = fip.encode('utf-8')
        except AttributeError:
            binary_fip_path = self.binary_fip_path
        rc = self._xlink.xlink_boot_device(byref(handle), binary_fip_path)
        return rc

    def boot_device(self, sw_id, fip=None):
        if platform.system() == "Windows":
            return self._win_boot_device(sw_id, fip)
        else:
            return self._linux_boot_device(sw_id, fip)

    def _sysfs_function(self, device, func_name, value="1"):
        fn = os.path.join(device.sysfs_path, func_name)
        try:
            with open(fn, "w") as f:
                f.write(value)
        except Exception as e:
            raise SysDeviceError(fn, e, self) from None

    def remove(self, device):
        self._sysfs_function(device, "remove")

    def _win_create_gpt(self, device):
        handle = XlinkHandle()
        handle.sw_device_id = c_uint32(device.sw_id)
        handle.dev_type = c_int(1)
        rc = self._xlink.xlink_create_partition(byref(handle))
        if rc != 0:
            raise XlinkAPIError('xlink.xlink_create_partition', rc)
        return rc

    def _linux_create_gpt(self, device):
        self._sysfs_function(device, "create_partitions")

    def create_gpt(self, device):
        if platform.system() == "Windows":
            return self._win_create_gpt(device)
        else:
            return self._linux_create_gpt(device)

    def _win_recovery_done(self, device):
        handle = XlinkHandle()
        handle.sw_device_id = c_uint32(device.sw_id)
        handle.dev_type = c_int(1)
        rc = self._xlink.xlink_recovery_done(byref(handle))
        if rc != 0:
            raise XlinkAPIError('xlink.xlink_recovery_done', rc)
        return rc

    def _linux_recovery_done(self, device):
        self._sysfs_function(device, "recovery_done")

    def recovery_done(self, device):
        if platform.system() == "Windows":
            return self._win_recovery_done(device)
        else:
            return self._linux_recovery_done(device)

    def _win_erase(self, device, part_name):
        raise NotImplementedError

    def _linux_erase(self, device, part_name):
        self._sysfs_function(device, "partition", part_name)
        self._sysfs_function(device, "erase_partition")

    def erase(self, device, part_name):
        if platform.system() == "Windows":
            return self._win_erase(device, part_name)
        else:
            return self._linux_erase(device, part_name)

    def _win_flash_partition(self, device, part_name, img):
        handle = XlinkHandle()
        handle.sw_device_id = c_uint32(device.sw_id)
        handle.dev_type = c_int(1)
        rc = self._xlink.xlink_flash_partition(byref(handle), part_name, img)
        if rc != 0:
            raise XlinkAPIError('xlink.flash_partition', rc)
        return rc

    def _linux_flash_partition(self, device, part_name, img):
        try:
            self._sysfs_function(device, "partition", part_name)
            fn = os.path.join(device.sysfs_path, "image_data")
            with open(img, 'rb') as src, open(fn, 'wb') as dst:
                while True:
                    chunk = src.read(1024 * 64)
                    if not chunk:
                        break
                    dst.write(chunk)
        except Exception as e:
            raise SysDeviceError(fn, e, self) from None

    def flash_partition(self, device, part_name, img):
        if platform.system() == "Windows":
            return self._win_flash_partition(device, part_name, img)
        else:
            return self._linux_flash_partition(device, part_name, img)
